from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Offsets:
    # in radians in FLU
    roll: Optional[float] = None
    pitch: Optional[float] = None
    yaw: Optional[float] = None
    forward: Optional[float] = None
    left: Optional[float] = None
    up: Optional[float] = None


@dataclass
class Camera:
    # This will become the name of frame-id, ros topic and webrtc stream
    name: str
    # Used to order the stream in the UI
    order: int
    # An array of gstream source / transform elements e.g.
    # ["v4l2src", "video/x-raw, format=RGB,width=1920,height=1080"]
    elements: List[str]

    pixel_width: int
    pixel_height: int
    sensor_width_mm: float
    sensor_height_mm: float
    fov: float
    # Published ros topic names
    camera_frame_topic: str
    camera_info_topic: str
    camera_info_ext_topic: str
    # Camera intrinsics and distortion parameters from callibration
    k_intrinsic: Optional[List[float]] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]
    distortion_parameters: Optional[List[float]] = [0.0, 0.0, 0.0, 0.0, 0.0]
    distortion_model: str = "plumb_bob"
    # The camera's position relative to the vessel base_link
    offsets: Optional[Offsets] = None
    type: str = "color"
    ros_throttle_time: float = 0.0000001
    # Whether to undistort the image in gstreamer pipeline before publishing
    undistort_image: bool = False


@dataclass
class CameraOverride:
    # This will become the name of frame-id, ros topic and webrtc stream
    name: Optional[str]
    order: Optional[int]
    # An array of gstream source / transform elements e.g.
    # ["v4l2src", "video/x-raw, format=RGB,width=1920,height=1080"]
    elements: Optional[List[str]]
    pixel_width: Optional[int]
    pixel_height: Optional[int]
    sensor_width_mm: Optional[float]
    sensor_height_mm: Optional[float]
    fov: Optional[float]
    # Published ros topic names
    camera_frame_topic: Optional[str]
    camera_info_topic: Optional[str]
    camera_info_ext_topic: Optional[str]
    # Camera intrinsics and distortion parameters from callibration
    distortion_parameters: Optional[List[float]] = None
    distortion_model: str = "plumb_bob"
    # The camera's position relative to the vessel base_link
    offsets: Optional[Offsets] = None
    type: Optional[str] = "color"
    ros_throttle_time: Optional[float] = 0.0000001
    # Whether to undistort the image in gstreamer pipeline before publishing
    undistort_image: Optional[bool] = False


@dataclass
class GreenstreamConfig:
    cameras: List[Camera]
    camera_overrides: Optional[List[Optional[CameraOverride]]] = None
    signalling_server_port: int = 8443
    namespace: str = ""  # eg. vessel_1
    ui_port: int = 8080
    mode: str = "simulator"
    variant: str = ""



from typing import List, Optional, Tuple

from urchin import Joint, Link, xyz_rpy_to_matrix

from greenstream_config.merge_cameras import merge_cameras
from greenstream_config.types import Camera, CameraOverride


def get_camera_urdf(
    camera: Camera, add_optical_frame: bool = True
) -> Tuple[List[Link], List[Joint]]:
    # This is the camera urdf from the gama/lookout greenstream.launch.py
    # We need to generate this from the camera config
    camera_xyz_rpy = (
        [
            camera.offsets.forward or 0.0,
            camera.offsets.left or 0.0,
            camera.offsets.up or 0.0,
            camera.offsets.roll or 0.0,
            camera.offsets.pitch or 0.0,
            camera.offsets.yaw or 0.0,
        ]
        if camera.offsets
        else [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )

    links = [Link(name=f"{camera.name}_link", inertial=None, visuals=None, collisions=None)]
    joints = [
        Joint(
            name=f"{camera.name}_joint",
            parent="base_link",
            child=f"{camera.name}_link",
            joint_type="fixed",
            origin=xyz_rpy_to_matrix(camera_xyz_rpy),
        )
    ]

    if add_optical_frame:
        links.append(
            Link(
                name=f"{camera.name}_{camera.type}_frame",
                inertial=None,
                visuals=None,
                collisions=None,
            )
        )
        links.append(
            Link(
                name=f"{camera.name}_{camera.type}_optical_frame",
                inertial=None,
                visuals=None,
                collisions=None,
            )
        )

        # fixed transforms between camera frame and optical frame FRD -> NED
        joints.append(
            Joint(
                name=f"{camera.name}_link_to_{camera.type}_frame",
                parent=f"{camera.name}_link",
                child=f"{camera.name}_{camera.type}_frame",
                joint_type="fixed",
                origin=xyz_rpy_to_matrix([0, 0, 0, 0, 0, 0]),
            )
        )
        joints.append(
            Joint(
                name=f"{camera.name}_{camera.type}_frame_to_optical_frame",
                parent=f"{camera.name}_{camera.type}_frame",
                child=f"{camera.name}_{camera.type}_optical_frame",
                joint_type="fixed",
                origin=xyz_rpy_to_matrix([0, 0, 0, -1.570796, 0, -1.570796]),
            )
        )

    return (links, joints)


def get_cameras_urdf(
    cameras: List[Camera],
    camera_overrides: List[Optional[CameraOverride]],
    add_optical_frame: bool = True,
) -> Tuple[List[Link], List[Joint]]:

    links: List[Link] = []
    joints: List[Joint] = []
    cameras = merge_cameras(cameras, camera_overrides)

    # assume cameras have already been merged with overrides
    for camera in cameras:
        camera_links, camera_joints = get_camera_urdf(camera, add_optical_frame)
        links += camera_links
        joints += camera_joints

    return links, joints



from greenstream_config.merge_cameras import merge_cameras
from greenstream_config.types import Camera, CameraOverride, GreenstreamConfig, Offsets
from greenstream_config.urdf import get_camera_urdf, get_cameras_urdf

__all__ = [
    "GreenstreamConfig",
    "Camera",
    "CameraOverride",
    "Offsets",
    "get_camera_urdf",
    "get_cameras_urdf",
    "merge_cameras",
]
